﻿' These four mutually-exclusive defines govern how the lines are rendered
' -----------------------------------------------------------------------

' #define USE_WRITEABLEBITMAP_WITH_POLYLINE
' #define USE_WRITEABLEBITMAP_WITH_OUTLINE_FILL
#Const USE_CANVAS_WITH_ANIMATION_FADEOUT = True
' #define USE_CANVAS_WITH_MANUAL_FADEOUT

#If USE_WRITEABLEBITMAP_WITH_POLYLINE OrElse USE_WRITEABLEBITMAP_WITH_OUTLINE_FILL Then
#Const USE_WRITEABLEBITMAP = True
#End If

#If USE_CANVAS_WITH_ANIMATION_FADEOUT OrElse USE_CANVAS_WITH_MANUAL_FADEOUT Then
#Const USE_CANVAS = True
#End If

#If USE_WRITEABLEBITMAP AndAlso USE_CANVAS Then
'#error Only one of four rendering symbols in Oscilloscope.xaml.vb can be defined!
#End If


Imports Microsoft.VisualBasic

Partial Public Class Oscilloscope
    Inherits UserControl
    Private lastDateTime As Date = Date.Now

#If USE_WRITEABLEBITMAP Then

        Private writeableBitmap As WriteableBitmap

#End If ' USE_WRITEABLEBITMAP

#If USE_WRITEABLEBITMAP_WITH_OUTLINE_FILL Then

        Private Const beamPixel As Integer = CInt(&HFF00FF00)
        Private points As New PointCollection
        Private lineList As New List(Of RoundCappedLine)
        Private xCollection As New List(Of Double)

#Else
    Private strokeBrush As New SolidColorBrush(Color.FromArgb(&HFF, 0, &HFF, 0))
#End If

#If Not USE_CANVAS_WITH_ANIMATION_FADEOUT Then

        Private accumulatedDecrease As Double

#End If '!USE_CANVAS_WITH_ANIMATION_FADEOUT

#If USE_CANVAS_WITH_MANUAL_FADEOUT Then

        Private removableElements As New List(Of UIElement)

#End If 'USE CANVAS_WITH_MANUAL_FADEOUT

    Public Shared ReadOnly XProviderProperty As DependencyProperty = DependencyProperty.Register("XProvider", GetType(IProvideAxisValue), GetType(Oscilloscope), New PropertyMetadata(Nothing))

    Public Shared ReadOnly YProviderProperty As DependencyProperty = DependencyProperty.Register("YProvider", GetType(IProvideAxisValue), GetType(Oscilloscope), New PropertyMetadata(Nothing))

    Public Shared ReadOnly BrightnessProperty As DependencyProperty = DependencyProperty.Register("Brightness", GetType(Double), GetType(Oscilloscope), New PropertyMetadata(2.0))

    Public Shared ReadOnly PersistenceProperty As DependencyProperty = DependencyProperty.Register("Persistence", GetType(Double), GetType(Oscilloscope), New PropertyMetadata(4.0))

#If USE_CANVAS_WITH_ANIMATION_FADEOUT Then

    Public Shared ReadOnly AnimatedObjectProperty As DependencyProperty = DependencyProperty.RegisterAttached("AnimatedObject", GetType(FrameworkElement), GetType(Oscilloscope), New PropertyMetadata(Nothing))

#End If 'USE_CANVAS_WITH_ANIMATION_FADEOUT

    Public Sub New()
        InitializeComponent()
    End Sub

    Public Property XProvider As IProvideAxisValue
        Set(ByVal value As IProvideAxisValue)
            SetValue(XProviderProperty, value)
        End Set
        Get
            Return CType(GetValue(XProviderProperty), IProvideAxisValue)
        End Get
    End Property

    Public Property YProvider As IProvideAxisValue
        Set(ByVal value As IProvideAxisValue)
            SetValue(YProviderProperty, value)
        End Set
        Get
            Return CType(GetValue(YProviderProperty), IProvideAxisValue)
        End Get
    End Property

    Public Property Brightness As Double
        Set(ByVal value As Double)
            SetValue(BrightnessProperty, value)
        End Set
        Get
            Return CDbl(GetValue(BrightnessProperty))
        End Get
    End Property

    Public Property Persistence As Double
        Set(ByVal value As Double)
            SetValue(PersistenceProperty, value)
        End Set
        Get
            Return CDbl(GetValue(PersistenceProperty))
        End Get
    End Property

    Private Sub OnLayoutRootSizeChanged(ByVal sender As Object, ByVal args As SizeChangedEventArgs)
        If Not args.NewSize.IsEmpty Then
            lineGrid.Children.Clear()
            Dim grayBrush As New SolidColorBrush(Colors.LightGray)

            ' Display vertical grid lines
            For x As Double = 0 To args.NewSize.Width Step args.NewSize.Width \ 10
                Dim line As Line =
                    New Line With {
                        .X1 = x,
                        .Y1 = 0,
                        .X2 = x,
                        .Y2 = args.NewSize.Height,
                        .Stroke = grayBrush,
                        .StrokeThickness = If(x = args.NewSize.Width \ 2, 2, 1)
                    }
                lineGrid.Children.Add(line)
            Next x

            'Display horizontal grid lines
            For y As Double = 0 To args.NewSize.Height Step args.NewSize.Height \ 8
                Dim line As Line =
                    New Line With {
                        .X1 = 0,
                        .Y1 = y,
                        .X2 = args.NewSize.Width,
                        .Y2 = y,
                        .Stroke = grayBrush,
                        .StrokeThickness = If(y = args.NewSize.Height \ 2, 2, 1)
                    }

                lineGrid.Children.Add(line)
            Next y

#If USE_WRITEABLEBITMAP Then

                'Create WriteableBitmap
                writeableBitmap = New WriteableBitmap(CInt(Fix(args.NewSize.Width)), CInt(Fix(args.NewSize.Height)))
                screenImage.Source = writeableBitmap
                screenImage.Visibility = Visibility.Visible
                screenCanvas.Visibility = Visibility.Collapsed

#End If ' USE_WRITEABLEBITMAP

            'Start up the Rendering event
            AddHandler CompositionTarget.Rendering, AddressOf OnCompositionTargetRendering
        Else
            RemoveHandler CompositionTarget.Rendering, AddressOf OnCompositionTargetRendering
        End If
    End Sub

    Private Sub OnCompositionTargetRendering(ByVal sender As Object, ByVal args As EventArgs)
        Dim dateTime As Date = Date.Now

#If Not USE_CANVAS_WITH_ANIMATION_FADEOUT Then

            'Determine the amount to decrease the intensity of each pixel
            accumulatedDecrease += 256 * (dateTime.Subtract(lastDateTime)).TotalSeconds / Persistence
            Dim decrease As Integer = CInt(Fix(accumulatedDecrease))

            'If integral decrease, sweep through the pixels
            If decrease > 0 Then
                accumulatedDecrease -= decrease

            #If USE_WRITEABLEBITMAP Then

                For index = 0 To writeableBitmap.Pixels.Length - 1
                    Dim pixel As Integer = writeableBitmap.Pixels(index)

                    If pixel <> 0 Then
                        Dim a As Integer = pixel >> 24 And &HFF
                        Dim r As Integer = pixel >> 16 And &HFF
                        Dim g As Integer = pixel >> 8 And &HFF
                        Dim b As Integer = pixel And &HFF

                        a = Math.Max(0, a - decrease)
                        r = Math.Max(0, r - decrease)
                        g = Math.Max(0, g - decrease)
                        b = Math.Max(0, b - decrease)

                        writeableBitmap.Pixels(index) = a << 24 Or r << 16 Or g << 8 Or b
                    End If
                Next index

            #ElseIf USE_CANVAS_WITH_MANUAL_FADEOUT Then

                'Decrease the color of each element
                For Each child As UIElement In screenCanvas.Children
                    Dim brush As SolidColorBrush = TryCast((TryCast(child, Shape)).Stroke, SolidColorBrush)
                    Dim clr As Color = brush.Color
                    brush.Color = Color.FromArgb(CByte(Math.Max(0, clr.A - decrease)), clr.R, clr.G, clr.B)

                    If brush.Color.A = 0 Then
                        removableElements.Add(child)
                    End If
                Next child

                For Each child As UIElement In removableElements
                    screenCanvas.Children.Remove(child)
                Next child
            #End If
            End If
#End If
        If XProvider IsNot Nothing AndAlso YProvider IsNot Nothing Then

#If USE_WRITEABLEBITMAP_WITH_OUTLINE_FILL Then

                points.Clear()

#Else
            Dim polyline = New Polyline With {.StrokeThickness = Me.Brightness,
                                              .Stroke = New SolidColorBrush(Color.FromArgb(&HFF, 0, &HFF, 0)),
                                              .StrokeStartLineCap = PenLineCap.Round,
                                              .StrokeEndLineCap = PenLineCap.Round,
                                              .StrokeLineJoin = PenLineJoin.Round}

            Dim points = polyline.Points
#End If

            'Interpolate from lastDateTime to current dateTime
            Dim milliseconds = CInt((dateTime.Subtract(lastDateTime).TotalMilliseconds))

            For msec = 0 To milliseconds
                Dim dt = lastDateTime + TimeSpan.FromMilliseconds(msec)

                Dim point As New Point(0.5 * LayoutRoot.ActualWidth * (1 + XProvider.GetAxisValue(dt)),
                                       0.5 * LayoutRoot.ActualHeight * (1 + YProvider.GetAxisValue(dt)))
                points.Add(point)
            Next msec

            lastDateTime = dateTime

#If USE_CANVAS_WITH_ANIMATION_FADEOUT Then

            Dim anima As ColorAnimation = New ColorAnimation With
                                          {.To = Color.FromArgb(0, 0, &HFF, 0),
                                           .Duration = TimeSpan.FromSeconds(Me.Persistence)}

            System.Windows.Media.Animation.Storyboard.SetTarget(anima, TryCast(polyline.Stroke, SolidColorBrush))
            System.Windows.Media.Animation.Storyboard.SetTargetProperty(anima, New PropertyPath(SolidColorBrush.ColorProperty))

            Dim storyboard As New Storyboard
            storyboard.SetValue(Oscilloscope.AnimatedObjectProperty, polyline)
            storyboard.Children.Add(anima)
            AddHandler storyboard.Completed, AddressOf storyboard_Completed
            storyboard.Begin()
#End If

#If USE_CANVAS Then

            screenCanvas.Children.Add(polyline)

#ElseIf USE_WRITEABLEBITMAP_WITH_POLYLINE Then

                writeableBitmap.Render(polyline, Nothing)

#Else 'USE_WRITEABLEBITMAP_WITH_OUTLINE_FILL

                lineList.Clear()

                For i As Integer = 0 To points.Count - 2
                    Dim line As New RoundCappedLine(points(i), points(i + 1), Me.Brightness / 2)
                    lineList.Add(line)
                Next i

                For y  = 0 To writeableBitmap.PixelHeight - 1
                    xCollection.Clear()

                    For Each line As RoundCappedLine In lineList
                        line.GetAllX(y, xCollection)
                    Next line

                    If xCollection.Count > 0 Then
                        Dim rowIndex As Integer = y * writeableBitmap.PixelWidth

                        For pair = 0 To xCollection.Count - 1 Step 2
                            Dim x1 As Integer = CInt(Fix(xCollection(pair)))
                            Dim x2 As Integer = CInt(Fix(xCollection(pair + 1)))

                            If x1 <> x2 Then
                                Dim xMin As Integer = Math.Max(0, Math.Min(writeableBitmap.PixelWidth - 1, Math.Min(x1, x2)))
                                Dim xMax As Integer = Math.Max(0, Math.Min(writeableBitmap.PixelWidth - 1, Math.Max(x1, x2)))

                                For x As Integer = xMin To xMax - 1
                                    writeableBitmap.Pixels(rowIndex + x) = beamPixel
                                Next x
                            End If
                        Next pair
                    End If
                Next y
#End If
        End If

#If USE_WRITEABLEBITMAP Then

            writeableBitmap.Invalidate()
#End If

    End Sub

#If USE_CANVAS_WITH_ANIMATION_FADEOUT Then

    Private Sub storyboard_Completed(ByVal sender As Object, ByVal e As EventArgs)
        Dim storyboard As Storyboard = TryCast(sender, Storyboard)
        Dim animatedObject As FrameworkElement = TryCast(storyboard.GetValue(Oscilloscope.AnimatedObjectProperty), FrameworkElement)
        screenCanvas.Children.Remove(animatedObject)
    End Sub

#End If

End Class